// Copyright (c) 2022 HTC Corporation. All Rights Reserved.


#include "AnimNode_FacialTrackingPose.h"
#include "FacialTrackingRemapAsset.h"
#include "Animation/AnimInstanceProxy.h"
#include "Animation/AnimTrace.h"

FAnimNode_FacialTrackingPose::FAnimNode_FacialTrackingPose()
: CachedDeltaTime(0.0f),
  RemapAsset(nullptr)
{
}

void FAnimNode_FacialTrackingPose::Initialize_AnyThread(const FAnimationInitializeContext& Context)
{
  InputPose.Initialize(Context);
}

void FAnimNode_FacialTrackingPose::CacheBones_AnyThread(const FAnimationCacheBonesContext& Context)
{
  Super::CacheBones_AnyThread(Context);
  InputPose.CacheBones(Context);
}

void FAnimNode_FacialTrackingPose::Update_AnyThread(const FAnimationUpdateContext& Context)
{
  InputPose.Update(Context);

  GetEvaluateGraphExposedInputs().Execute(Context);

  // Accumulate Delta time from update
  CachedDeltaTime += Context.GetDeltaTime();

#if ANIM_TRACE_ENABLED
  if (EyeBlendShapeMap.Num() > 0)
  {
    for (auto eyeShapes : EyeBlendShapeMap)
    {
      FString eyeShapeName = *(RemapAsset->GetEnumValueAsString<EEyeShape>("EEyeShape", eyeShapes.Key));
      TRACE_ANIM_NODE_VALUE(Context, TEXT("Eye shape"), *eyeShapeName);
      TRACE_ANIM_NODE_VALUE(Context, TEXT("Eye weighting"), eyeShapes.Value);
    }
  }

  if (LipBlendShapeMap.Num() > 0)
  {
    for (auto lipShapes : LipBlendShapeMap)
    {
      FString lipShapeName = *(RemapAsset->GetEnumValueAsString<ELipShape_v2>("ELipShape_v2", lipShapes.Key));
      TRACE_ANIM_NODE_VALUE(Context, TEXT("Lip shape"), *lipShapeName);
      TRACE_ANIM_NODE_VALUE(Context, TEXT("Lip weighting"), lipShapes.Value);
    }
  }
#endif // ANIM_TRACE_ENABLED
}

void FAnimNode_FacialTrackingPose::Evaluate_AnyThread(FPoseContext& Output)
{
  InputPose.Evaluate(Output);
  if (EyeBlendShapeMap.Num() > 0 || LipBlendShapeMap.Num() > 0)
  {
    RemapAsset->BuildPoseAndCurveFromBaseData(CachedDeltaTime, EyeBlendShapeMap, LipBlendShapeMap, Output.Pose, Output.Curve);
    CachedDeltaTime = 0.f;
  }
}

void FAnimNode_FacialTrackingPose::PreUpdate(const UAnimInstance* InAnimInstance)
{
  RemapAsset = NewObject<UFacialTrackingRemapAsset>();
}

void FAnimNode_FacialTrackingPose::GatherDebugData(FNodeDebugData& DebugData)
{
  const int NumEyeShapes = EyeBlendShapeMap.Num();
  const int NumLipShapes = LipBlendShapeMap.Num();
  FString DebugLine = FString::Printf(TEXT("Num Eye Shapes: %i, Num Lip Shapes: %i."), NumEyeShapes, NumLipShapes);

  DebugData.AddDebugItem(DebugLine);
  InputPose.GatherDebugData(DebugData);
}

void FAnimNode_FacialTrackingPose::OnInitializeAnimInstance(const FAnimInstanceProxy* InProxy, const UAnimInstance* InAnimInstance)
{
  Super::OnInitializeAnimInstance(InProxy, InAnimInstance);
}
