// Copyright (c) 2022 HTC Corporation. All Rights Reserved.


#include "FacialTrackingRemapAsset.h"
#include "Animation/AnimCurveTypes.h"
#include "Animation/AnimTypes.h"
#include "Animation/Skeleton.h"
#include "BonePose.h"
#include "GenericPlatform/GenericPlatformMath.h"

UFacialTrackingRemapAsset::UFacialTrackingRemapAsset(const FObjectInitializer& objectInitializer)
  : Super(objectInitializer)
{
}


void UFacialTrackingRemapAsset::ApplyCurveValue(const USkeleton* Skeleton, const FName CurveName, const float CurveValue, FBlendedCurve& OutCurve) const
{
  SmartName::UID_Type UID = Skeleton->GetUIDByName(USkeleton::AnimCurveMappingName, CurveName);
  if (UID != SmartName::MaxUID)
  {
    OutCurve.Set(UID, CurveValue);
  }
}

void UFacialTrackingRemapAsset::BuildCurveData(const TMap<FName, float>& CurveMap, const FCompactPose& InPose, FBlendedCurve& OutCurve) const
{
  const USkeleton* Skeleton = InPose.GetBoneContainer().GetSkeletonAsset();

  for (const TPair<FName, float>& Pair : CurveMap)
  {
    ApplyCurveValue(Skeleton, Pair.Key, Pair.Value, OutCurve);
  }
}

void UFacialTrackingRemapAsset::BuildPoseAndCurveFromBaseData(float DeltaTime, const TMap<EEyeShape, float> EyeBlendShapes, const TMap<ELipShape_v2, float> LipBlendShapes, FCompactPose& OutPose, FBlendedCurve& OutCurve)
{
  TMap<FName, float> curveValuesMap;

  if (EyeBlendShapes.Num() > 0)
  {
    for (const auto eye : EyeBlendShapes)
    {      
      curveValuesMap.Emplace(*(GetEnumValueAsString<EEyeShape>("EEyeShape", eye.Key)), eye.Value);
    }
  }

  if (LipBlendShapes.Num() > 0)
  {
    for (const auto lip : LipBlendShapes)
    {
      curveValuesMap.Emplace(*(GetEnumValueAsString<ELipShape_v2>("ELipShape_v2", lip.Key)), lip.Value);
    }
  }
  BuildCurveData(curveValuesMap, OutPose, OutCurve);
}
