// Copyright (c) 2022 HTC Corporation. All Rights Reserved.

#pragma once
#include "Animation/AnimNodeBase.h"
#include "CoreMinimal.h"
#include "FacialExpressionEnums.h"
#include "FacialTrackingRemapAsset.h"
#include "AnimNode_FacialTrackingPose.generated.h"

class UFacialTrackingRemapAsset;

PRAGMA_DISABLE_DEPRECATION_WARNINGS
USTRUCT(BlueprintInternalUseOnly)
struct VIVEOPENXR_API FAnimNode_FacialTrackingPose : public FAnimNode_Base
{
	GENERATED_BODY()

public:
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = Input)
	FPoseLink InputPose;

	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = SourceData, meta = (PinShownByDefault))
	TMap<EEyeShape, float> EyeBlendShapeMap;

	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = SourceData, meta = (PinShownByDefault))
	TMap<ELipShape_v2, float> LipBlendShapeMap;

	FAnimNode_FacialTrackingPose();

	//~ FAnimNode_Base interface

	// Called whenever we need to Initialize/Reinitialize (when changing mesh for instance).
	virtual void Initialize_AnyThread(const FAnimationInitializeContext& Context) override;

	// For update all skeletal info.
	virtual void CacheBones_AnyThread(const FAnimationCacheBonesContext & Context) override;	

	// Called to update current state (such as advancing play time or updating blend weights). 
	// This function takes an FAnimationUpdateContext that knows the DeltaTime for the update and the current nodes blend weight.
	virtual void Update_AnyThread(const FAnimationUpdateContext & Context) override;	

	// Called to update current state (such as advancing play time or updating blend weights).
	// This function takes an FAnimationUpdateContext that knows the DeltaTime for the update and the current nodes blend weight.
	// Caculate All anim skeletal transform info , in the end it will Output a Pose to render thread and save in local component.
	virtual void Evaluate_AnyThread(FPoseContext& Output) override;									
	virtual bool HasPreUpdate() const { return true; }
	virtual void PreUpdate(const UAnimInstance* InAnimInstance) override;
	virtual void GatherDebugData(FNodeDebugData& DebugData) override;
	//~ End of FAnimNode_Base interface
protected:
	virtual void OnInitializeAnimInstance(const FAnimInstanceProxy* InProxy, const UAnimInstance* InAnimInstance) override;

	// Delta time from update so that it can be passed to retargeter
	float CachedDeltaTime;

	UFacialTrackingRemapAsset* RemapAsset;
};
PRAGMA_ENABLE_DEPRECATION_WARNINGS