// Copyright (c) 2022 HTC Corporation. All Rights Reserved.

#pragma once

#include "OpenXRCommon.h"
#include "CoreMinimal.h"
#include "ViveOpenXR.h"
#include "FacialExpressionEnums.h"

/**
 * 
 */
namespace ViveOpenXR
{
	enum class EError
	{
		FAILED = -1,
		WORK = 0
	};

	struct FFacialTrackingExtesionDispatchTable
	{
		PFN_xrGetSystemProperties xrGetSystemProperties;
		PFN_xrCreateFacialTrackerHTC xrCreateFacialTrackerHTC;
		PFN_xrDestroyFacialTrackerHTC xrDestroyFacialTrackerHTC;
		PFN_xrGetFacialExpressionsHTC xrGetFacialExpressionsHTC;
	};

	class FFacialTracking : public IOpenXRExtensionPlugin
	{
	public:
		FFacialTracking();
		~FFacialTracking();

		void Register();
		void Unregister();

		/** IOpenXRExtensionPlugin implementation */
		virtual const void* OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext) override;
		virtual const void* OnBeginSession(XrSession InSession, const void* InNext) override;
		virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
		virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;
		virtual void PostGetSystem(XrInstance InInstance, XrSystemId InSystem) override;
		virtual void PostCreateInstance(XrInstance InInstance) override;
		virtual void PostCreateSession(XrSession InSession) override;

		bool m_bEnableFacialTracking = false;

		bool CreateFacialTracker(XrFacialTrackingTypeHTC facialTrckingType);
		bool DestroyFacialTracker();
		bool GetEyeExpressions(bool& isActive, TMap<EEyeShape, float>& shapes);
		bool GetLipExpressions(bool& isActive, TMap<ELipShape_v2, float>& shapes);

		XrFacialTrackerHTC m_FacialTracker_Eye = XR_NULL_HANDLE;
		XrFacialTrackerHTC m_FacialTracker_Lip = XR_NULL_HANDLE;
	private:
		XrInstance m_XrInstance;
		XrSystemId m_XrSystemId;
		XrSession m_Session = XR_NULL_HANDLE;
		FFacialTrackingExtesionDispatchTable FT_ext{};

		const int eyeExpressionCount = ((int)EEyeShape::Max - 1); // (Max - 1) is because SRanipalAvatar have 15 shapes, and OpenXR EyeExpression shape count only 14.
		TMap<XrEyeExpressionHTC, float> eyeWeightings;
		TMap<EEyeShape, XrEyeExpressionHTC> eyeShapeMap;

		const int lipExpressionCount = (int)ELipShape_v2::Max;
		TMap<XrLipExpressionHTC, float> lipWeightings;
		TMap<ELipShape_v2, XrLipExpressionHTC> lipShapeMap;

		EError EyeLastUpdateResult = EError::FAILED;
		EError LipLastUpdateResult = EError::FAILED;
		XrFacialTrackingTypeHTC m_FacialTrackingType;

		double EyeLastUpdateFrame = -1;
		double LipLastUpdateFrame = -1;

		// Map OpenXR eye shapes to Avater eye blendshapes.
		void MapEyeShapes();
		void MapLipShapes();
	};
}