// Copyright (c) 2022 HTC Corporation. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "UObject/ObjectMacros.h"
#include "UObject/Object.h"
#include "FacialExpressionEnums.h"
#include "ViveOpenXR.h"
#include "FacialTrackingRemapAsset.generated.h"

class USkeleton;
struct FCompactPose;
struct FBlendedCurve;

UCLASS()
class VIVEOPENXR_API UFacialTrackingRemapAsset : public UObject
{
	GENERATED_UCLASS_BODY()
	
public:
	// Take the supplied curve name and value and applies it to the blended curve (as approriate given the supplied skeleton).
	void ApplyCurveValue(const USkeleton* Skeleton, const FName CurveName, const float CurveValue, FBlendedCurve& OutCurve) const;
	
	// Build curve data into OutCurve from the supplied map of curve name to float.
	void BuildCurveData(const TMap<FName, float>& CurveMap, const FCompactPose& InPose, FBlendedCurve& OutCurve) const;

	// Build OutPose and OutCurve from the basic data. Called for every type of subjects.
	void BuildPoseAndCurveFromBaseData(float DeltaTime, const TMap<EEyeShape, float> EyeBlendShapes, const TMap<ELipShape_v2, float> LipBlendShapes, FCompactPose& OutPose, FBlendedCurve& OutCurve);

	template<typename TEnum>
	static FORCEINLINE FString GetEnumValueAsString(const FString& EnumName, TEnum EnumValue)
	{
		const UEnum* EnumPtr = FindObject<UEnum>(ANY_PACKAGE, *EnumName, true);
		if (!EnumPtr) return FString("InValid");
		return EnumPtr->GetNameStringByIndex(static_cast<int32>(EnumValue));
	}
};
